 (*  Title:      Config.thy
    ID:         $Id: contexts.thy,v 1.4 2009/03/24 17:08:00 flokam Exp $
    Author:     Ludovic Henrio and Florian Kammuller
                2006

    Note:       Configurations, i.e. parallel activities for a functional
                ASP, based on the sigma-calculus and basic activities 
                defined in ASP.thy
                
                one lemma became an axiom to accelerate the computation time ... proof should work
                
*)

header {* Configurations for ASP *}

theory contexts imports Sigma begin

(* We first define atype for paires of (finite map from labels to dB) 
         x a label of this finite map *)
typedef FmapLabel = "{f::(Label -~> sterm) * Label . snd f \<in> dom (fst f)}"
  by auto

(* accessors *)
constdefs
     FLmap:: "FmapLabel \<Rightarrow> (Label -~> sterm)"
     "FLmap == fst\<circ>Rep_FmapLabel"
constdefs
     FLlabel:: "FmapLabel \<Rightarrow> Label"
     "FLlabel == snd\<circ>Rep_FmapLabel"

(* the term that corresponds to the label  in the Fmap*)
constdefs
     FLmapsto::"FmapLabel \<Rightarrow> sterm"
     "FLmapsto f == the ((FLmap f) (FLlabel f))"

lemma FLdef2: "(FLmap f) (FLlabel f) \<noteq> None"
proof -
  fix f :: FmapLabel
  have "(FLlabel f) \<in> dom (FLmap f)"
    unfolding FLmap_def FLlabel_def
  proof (simp add: FmapLabel_def Rep_FmapLabel_inject)
    from Rep_FmapLabel[of f] FmapLabel_def
    show "snd (Rep_FmapLabel f) \<in> dom (fst (Rep_FmapLabel f))"
      by auto
  qed
  thus "FLmap f (FLlabel f) \<noteq> None" by force
qed

lemma FLlabel_in_dom[simp]: "FLlabel f \<in> dom (FLmap f)"
  by (subgoal_tac "FLmap f (FLlabel f) \<noteq> None", force, rule FLdef2)

lemma FLequals: "\<lbrakk> \<forall>l. FLmap f l = FLmap f' l; FLlabel f = FLlabel f' \<rbrakk>
       \<Longrightarrow> f = f'"
proof -
  fix f :: FmapLabel and f' :: FmapLabel
  have "f' = Abs_FmapLabel (FLmap f', FLlabel f')"
    unfolding FLmap_def FLlabel_def
    by (force simp: Rep_FmapLabel_inverse)
  moreover
  have "f = Abs_FmapLabel (FLmap f, FLlabel f)"
    unfolding FLmap_def FLlabel_def
    by (simp add: Rep_FmapLabel_inverse)
  moreover
  assume "\<forall>l. FLmap f l = FLmap f' l"
  hence "FLmap f = FLmap f'" by (simp add: expand_fun_eq)
  moreover assume "FLlabel f = FLlabel f'"
  ultimately
  show "f = f'" by force
qed

lemma FLlabelAbs[simp]: "l \<in> dom f \<Longrightarrow> FLlabel(Abs_FmapLabel(f,l)) = l"
proof (simp add: FLlabel_def)
  fix l :: Label and f :: "Label -~> sterm"
  assume "l \<in> dom f"
  with Abs_FmapLabel_inverse[of "(f,l)"]
  show "snd (Rep_FmapLabel(Abs_FmapLabel(f,l))) = l"
    unfolding FmapLabel_def by simp
qed

lemma FLmapAbs[simp]: "l \<in> dom f \<Longrightarrow> FLmap(Abs_FmapLabel(f,l)) = f"
proof (simp add: FLmap_def)
  fix l :: Label and f :: "Label -~> sterm"
  assume "l \<in> dom f"
  with Abs_FmapLabel_inverse[of "(f,l)"]
  show "fst (Rep_FmapLabel(Abs_FmapLabel (f,l))) = f"
    unfolding FmapLabel_def by simp
qed

(* a general context is a term with a hole somewhere *)
datatype general_context =
   cHole
 | cObj FmapLabel type general_context
 | cCallL general_context Label sterm 
 | cCallR sterm Label general_context 
 | cUpdL general_context Label sterm
 | cUpdR sterm Label general_context
 | cActive general_context;

(* Fill fills the hole in the general context with a term given in parameter *)
primrec   
       Fill :: "[general_context, sterm]  \<Rightarrow>  sterm" (infixr "\<up>" 65)
where
  "Fill cHole x = x"
| "Fill (cObj f T E) x = Obj ((FLmap f)((FLlabel f) \<mapsto> (Fill E x))) T"
| "Fill (cCallL E l y) x = Call (Fill E x) l y"
| "Fill (cCallR y l E) x = Call y l(Fill E x)"
| "Fill (cUpdL E l y) x = Upd (Fill E x) l y"
| "Fill (cUpdR y l E) x = Upd y l (Fill E x)" 
| "Fill (cActive E) x = Active (Fill E x)"

(* econtexts are terms with a hole place in the leftmost position ("classical" evaluation context) *)
(* they are useful because the term placed in the hole o a closed term  has no free variable *)
datatype econtext = (*evaluation context*)
   eHole
 | eCall econtext Label sterm
 | eUpdL econtext Label sterm
 | eActive econtext 

(* eFill is the equivalent of Fill for econtexts *)
primrec eFill :: "[econtext, sterm]  \<Rightarrow>  sterm" (infixr "\<up>" 65)
where
  "eFill eHole x = x"
| "eFill (eCall E l b) x = Call (eFill E x) l b"
| "eFill (eUpdL E l y) x = Upd (eFill E x) l y"
| "eFill (eActive E) x = Active (eFill E x)"
 

(* cCompose is the composition of contexts: it places the second ctxt in the 
hole of the first ctxtx, this provides a new context *)
primrec cCompose :: "[general_context, general_context]  \<Rightarrow>  general_context" (infixr "\<circ>" 64)
where
  "cCompose cHole F = F"
| "cCompose (cObj f T E) F = cObj f T (cCompose E F)"
| "cCompose (cCallL E l x) F = cCallL (cCompose E F) l x"
| "cCompose (cCallR x l E) F = cCallR x l (cCompose E F)"
| "cCompose (cUpdL E l y) F = cUpdL (cCompose E F) l y"
| "cCompose (cUpdR y l E) F = cUpdR y l (cCompose E F)" 
| "cCompose (cActive E) F = cActive (cCompose E F)"

primrec eCompose :: "[econtext, econtext]  \<Rightarrow>  econtext" (infixr "\<circ>" 64)
where
  "eCompose eHole F = F"
| "eCompose (eCall E l x) F = eCall (eCompose E F) l x"
| "eCompose (eUpdL E l y) F = eUpdL (eCompose E F) l y"
| "eCompose (eActive E) F = eActive (eCompose E F)"

(* econtexts are general contexts, the following function does the conversion *)
primrec ectxt2gctxt:: " econtext \<Rightarrow> general_context"
where
  "ectxt2gctxt eHole = cHole"
| "ectxt2gctxt (eCall E l a) = cCallL (ectxt2gctxt E) l a"
| "ectxt2gctxt (eUpdL E l a) = cUpdL (ectxt2gctxt E) l a"
| "ectxt2gctxt (eActive E) = cActive (ectxt2gctxt E)"

section {*csize*}
(* the size of a context -- similar to size for terms; a hole ahas size 0 *) 
primrec csize:: "general_context \<Rightarrow> nat"
where
  "csize cHole = 0"
| "csize (cObj f T E) = ssize (Obj ((FLmap f)(FLlabel f \<mapsto> (Bvar (Self 0))) ) T) + csize E" 
    (* (Bvar (Self 0)) is of size 0 => we just want the size of the map with a hole*)
| "csize (cCallL E l x) = (csize E) + (ssize x) + Suc 0"
| "csize (cCallR x l E) = (csize E) + (ssize x) + Suc 0"
| "csize (cUpdL E l y) = (csize E) + (ssize y) + Suc 0"
| "csize (cUpdR y l E) = (ssize y) + (csize E ) + Suc 0" 
| "csize (cActive E)  = (csize E) + Suc 0"

primrec esize:: "econtext \<Rightarrow> nat"
where
  "esize eHole = 0"
| "esize (eCall E l x) = (esize E) + (ssize x)+ Suc 0"
| "esize (eUpdL E l y) = (esize E) + (ssize y) + Suc 0"
| "esize (eActive E) = (esize E) + Suc 0"

lemma "csize E \<noteq> csize F \<Longrightarrow> E \<noteq> F"
  by auto

lemma fold_cong: 
  "\<lbrakk> finite A; \<forall>x \<in> A . f x = g x \<rbrakk> 
  \<Longrightarrow> fold (\<lambda> x y . (y::nat) + (f x)) z A = fold  (\<lambda> x y . y + (g x)) z A"
proof(induct rule: finite.induct)
  show "fold (\<lambda>x y. y + f x) z {} = fold (\<lambda>x y. y + g x) z {}" by simp
next
  case (insertI A a) thus ?case
  proof (cases "a \<in> A")
    case True
    from prems have "(insert a A) = A"  by force
    thus ?thesis using prems by force
  next
    case False 
    from prems show ?thesis by simp
  qed
qed

(* relates size of terms and field update *)
lemma ssizeAndReplace: 
  "ssize (Obj ((FLmap f)(FLlabel f \<mapsto> b)) T) 
  = ssize (Obj ((FLmap f)(FLlabel f \<mapsto> a)) T) + ssize b - ssize a"
proof (auto)
  from finite_dom_fmap[of "FLmap f"]
  have
    "finite (dom (FLmap f) - {FLlabel f})"
    "\<forall>x\<in>dom (FLmap f) - {FLlabel f}.
      ssize_option (if x = FLlabel f then Some b else FLmap f x) 
      = ssize_option (if x = FLlabel f then Some a else FLmap f x)" 
    "FLlabel f \<in> dom (FLmap f)" by simp+
  from fold_cong[OF this(1-2)]
       insert_absorb[OF this(3)]
       fold_rec[OF finite_dom_fmap[of "FLmap f"] this(3)]
  show "fold (\<lambda>x y. y + ssize_option (if x = FLlabel f then Some b else FLmap f x))
         (Suc 0) (insert (FLlabel f) (dom (FLmap f))) 
        = fold (\<lambda>x y. y + ssize_option (if x = FLlabel f then Some a else FLmap f x))
           (Suc 0) (insert (FLlabel f) (dom (FLmap f))) + ssize b - ssize a"
    by simp
qed

(* size of filled ctxt is the sum of sizes of ctxt and term *)
lemma csize_and_Fill: "ssize ((E::general_context)\<up>x) = csize E + ssize x"
proof (induct E, simp_all)
  case (cObj f T E) 
  with ssizeAndReplace[of f "E\<up>x" _ "Bvar (Self 0)"]
  show "fold (\<lambda>xa y. y + ssize_option (if xa = FLlabel f then Some (E\<up>x) 
                                       else FLmap f xa))
         (Suc 0) (insert (FLlabel f) (dom (FLmap f))) 
        = fold (\<lambda>x y. y + ssize_option (if x = FLlabel f then Some (Bvar (Self 0))
                                        else FLmap f x))
          (Suc 0) (insert (FLlabel f) (dom (FLmap f))) + csize E + ssize x"
    by simp
qed

section {*Lemmas on contexts*}

lemma ectxt2gctxt_and_fill: "(ectxt2gctxt E)\<up>x = E\<up>x"
  by (induct E, auto)

lemma cCompose_and_Fill: "\<forall>F x. ((E::general_context)\<circ>F)\<up>x = E\<up>(F\<up>x)"
  by (induct E, auto)

lemma eCompose_and_Fill: "\<forall>F x. ((E::econtext)\<circ>F)\<up>x = E\<up>(F\<up>x)"
  by (induct E, auto)

lemma size_cHole: "csize (E::general_context)= 0 \<Longrightarrow> E = cHole"
proof (cases E)
  case cHole thus "E = cHole" by assumption
next
  case cCallL moreover assume "csize E = 0"
  ultimately show "E = cHole" by auto
next
  case cCallR moreover assume "csize E = 0"
  ultimately show "E = cHole" by auto
next
  case cUpdL moreover assume "csize E = 0"
  ultimately show "E = cHole" by auto
next
  case cUpdR moreover assume "csize E = 0"
  ultimately show "E = cHole" by auto
next
  case cActive moreover assume "csize E = 0"
  ultimately show "E = cHole" by auto
next
  case (cObj f T Ea)
  hence "ssize (E\<up>Bvar (Self 0)) > 0" 
    by (simp only: Fill.simps SizeOfObjectPos)
  moreover assume "csize E = 0"
  hence "ssize (E\<up>Bvar (Self 0)) = 0" 
    by (simp only: csize_and_Fill ssize_ssize_option.simps)
  ultimately
  show "E = cHole" by simp
qed

lemma fixpoint_context: "x = (E::general_context)\<up>x \<Longrightarrow> E = cHole"
proof -
  assume "x = E\<up>x" with csize_and_Fill[of E x]
  have "csize E = 0" by simp
  from size_cHole[OF this] show "E = cHole" by assumption
qed

lemma eq_context[rule_format]: "\<forall>x y. (E::general_context)\<up>x = E\<up>y \<longrightarrow> x = y"
proof (induct E, auto)
  case (cObj f T E)
  fix f :: FmapLabel and x :: sterm and y :: sterm
  assume "FLmap f(FLlabel f \<mapsto> E\<up>x) = FLmap f(FLlabel f \<mapsto> E\<up>y)"
  hence "(FLmap f(FLlabel f \<mapsto> E\<up>x)) (FLlabel f) = (FLmap f(FLlabel f \<mapsto> E\<up>y)) (FLlabel f)"
    by auto
  with prems(1) show "x = y" by auto
qed

(* equivalence on t=context is equality, except that the case of cObj because the term corresponding to the label can be anything *)
primrec equiv_contexts:: "general_context \<Rightarrow> general_context \<Rightarrow> bool" (infix "\<approx>" 60)
where
  "equiv_contexts cHole F = (F = cHole)"
| "equiv_contexts (cObj f T E) F = (\<exists>f' F'. F = cObj f' T F' \<and> E \<approx> F'
                                          \<and> FLlabel f = FLlabel f'
                                          \<and> (\<forall>l. l \<noteq> FLlabel f 
                                              \<longrightarrow> (FLmap f) l = (FLmap f') l))"
| "equiv_contexts (cCallL E l x) F = (\<exists>F'. F = cCallL F' l x \<and> E \<approx> F')"
| "equiv_contexts (cCallR x l E) F = (\<exists>F'. F = cCallR x l F' \<and> E \<approx> F')"
| "equiv_contexts (cUpdL E l y) F = (\<exists>F'. F = cUpdL F' l y \<and> E \<approx> F')"
| "equiv_contexts (cUpdR y l E) F = (\<exists>F'. F = cUpdR y l F' \<and> E \<approx> F')" 
| "equiv_contexts (cActive E) F = (\<exists>F'. F = cActive F' \<and> E \<approx> F')"

lemma equiv_and_equals[rule_format]: "\<forall>E'. E = E' \<longrightarrow> E \<approx> E'"
  by (induct E, auto)

lemma equiv_and_fill[rule_format]: "\<forall>x E'. E \<approx> E' \<longrightarrow> E\<up>x = E'\<up>x"
proof (induct E, auto)
  case (cObj f T E)
  fix f' :: FmapLabel and F' :: general_context and x :: sterm
  assume "FLlabel f = FLlabel f'"
         "\<forall>l. l \<noteq> FLlabel f' \<longrightarrow> FLmap f l = FLmap f' l"
  with expand_fun_eq[of "FLmap f(FLlabel f' \<mapsto> F'\<up>x)"
                        "FLmap f'(FLlabel f' \<mapsto> F'\<up>x)"]
  show "FLmap f(FLlabel f' \<mapsto> F'\<up>x) = FLmap f'(FLlabel f' \<mapsto> F'\<up>x)"
    by auto
qed

lemma gc_obj_access_filled:
 "(FLmap f(FLlabel f \<mapsto> (E::general_context)\<up>x)) (FLlabel f) =  Some (E\<up>x)"
  by auto

lemma gcexists_lemma0: 
  fixes E :: general_context and Ea :: general_context and 
        f :: FmapLabel and fa :: FmapLabel and
        x :: sterm and y :: sterm
  assumes 
  "\<forall>x y (E'::general_context). (x \<noteq> y \<and> (E::general_context)\<up>x = E'\<up>x \<and> E\<up>y = E'\<up>y) 
    \<longrightarrow> E \<approx> E'"
  "x \<noteq> y" "FLmap f(FLlabel f \<mapsto> E\<up>x) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>x)"
  "FLmap f(FLlabel f \<mapsto> E \<up> y) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)"
  shows "FLlabel f = FLlabel fa"
proof (cases "FLlabel f = FLlabel fa")
  case True thus ?thesis by blast
next
  case False
  from `FLmap f(FLlabel f \<mapsto> E\<up>x) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>x)`
       gc_obj_access_filled[of f E x] `FLlabel f \<noteq> FLlabel fa`
  have "Some (E\<up>x) = FLmap fa (FLlabel f)" 
    by (simp add: fun_upd_def)
  moreover
  from `FLmap f(FLlabel f \<mapsto> E\<up>y) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)`
       gc_obj_access_filled[of f E y] `FLlabel f \<noteq> FLlabel fa`
  have "FLmap fa (FLlabel f) = Some (E\<up>y)" 
    by (simp add: fun_upd_def)
  ultimately
  have "E\<up>x = E\<up>y" by simp
  from eq_context[OF this] `x \<noteq> y`
  show ?thesis by simp
qed

lemma general_context_exists:
 "\<forall>x y (E'::general_context). x \<noteq> y \<and> (E::general_context)\<up>x = E'\<up>x \<and> E\<up>y = E'\<up>y
    \<longrightarrow> E \<approx> E'"
proof (induct E)
  case cHole thus ?case by (auto simp: fixpoint_context)
next
  case (cObj f T E) show ?case
  proof (simp, intro strip, elim conjE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "x \<noteq> y"
           "Obj (FLmap f(FLlabel f \<mapsto> E\<up>x)) T = E'\<up>x"
           "Obj (FLmap f(FLlabel f \<mapsto> E\<up>y)) T = E'\<up>y"
    thus "\<exists>f' F'. E' = cObj f' T F' 
	        \<and> E \<approx> F' 
	        \<and> FLlabel f = FLlabel f' 
	        \<and> (\<forall>l. l \<noteq> FLlabel f \<longrightarrow> FLmap f l = FLmap f' l)"
    proof (cases E', auto simp: fixpoint_context)
      assume "Obj (FLmap f(FLlabel f \<mapsto> E\<up>x)) T = x"
      hence "x = (cObj f T E)\<up>x" by auto
      from fixpoint_context[OF this] show False by simp
    next
      fix fa :: FmapLabel and Ea :: general_context
      assume "FLmap f(FLlabel f \<mapsto> E\<up>x) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>x)"
	     "FLmap f(FLlabel f \<mapsto> E\<up>y) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)"
      from this(1) gcexists_lemma0[OF prems(1) `x \<noteq> y` this]
	   gc_obj_access_filled[of f E x] 
	   gc_obj_access_filled[of fa Ea x] 
      have "E\<up>x = Ea\<up>x" by simp
      moreover
      from `FLmap f(FLlabel f \<mapsto> E\<up>y) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)`
	   gcexists_lemma0[OF prems(1) `x \<noteq> y`
                              `FLmap f(FLlabel f \<mapsto> E\<up>x) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>x)`
                              `FLmap f(FLlabel f \<mapsto> E\<up>y) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)`]
	   gc_obj_access_filled[of f E y] 
	   gc_obj_access_filled[of fa Ea y] 
      have "E\<up>y = Ea\<up>y" by simp
      ultimately
      show "E \<approx> Ea" using prems(1) `x \<noteq> y` by blast
    next
      fix fa :: FmapLabel and Ea :: general_context
      assume "FLmap f(FLlabel f \<mapsto> E\<up>x) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>x)"
	     "FLmap f(FLlabel f \<mapsto> E\<up>y) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)"
      from gcexists_lemma0[OF prems(1) `x \<noteq> y` this] 
      show "FLlabel f = FLlabel fa" by assumption
    next
      fix fa :: FmapLabel and Ea :: general_context and l :: Label
      assume "FLmap f(FLlabel f \<mapsto> E\<up>x) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>x)"
	     "FLmap f(FLlabel f \<mapsto> E\<up>y) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)"
	     "l \<noteq> FLlabel f"
      from gcexists_lemma0[OF prems(1) `x \<noteq> y` this(1-2)] this(3)
      have "l \<noteq> FLlabel fa" by simp
      from fun_upd_other[OF this]
      have "(FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)) l = FLmap fa l"  by fastsimp
      moreover
      from fun_upd_other[OF `l \<noteq> FLlabel f`]
      have "FLmap f l = (FLmap f(FLlabel f \<mapsto> E\<up>y)) l" by fastsimp
      ultimately
      show "FLmap f l = FLmap fa l" 
	using `FLmap f(FLlabel f \<mapsto> E\<up>y) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)` 
	by simp
    qed
  qed
next
  case (cCallL E l t) show ?case
  proof (simp, intro strip, elim conjE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "x \<noteq> y"
           "Call (E\<up>x) l t = E'\<up>x"
           "Call (E\<up>y) l t = E'\<up>y"
    thus "\<exists>F'. E' = cCallL F' l t \<and> E \<approx> F'" 
      using prems(1)
    proof (cases E', auto)
      assume "Call (E\<up>x) l t = x" hence "x = (cCallL E l t)\<up>x" by auto
      from fixpoint_context[OF this] show False by simp
    next
      assume "E\<up>y = E\<up>x"
      from eq_context[OF this] `x \<noteq> y` show False by simp
    qed
  qed
next
  case (cCallR t l E) show ?case
  proof (simp, intro strip, elim conjE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "x \<noteq> y"
           "Call t l (E\<up>x) = E'\<up>x"
           "Call t l (E\<up>y) = E'\<up>y"
    thus "\<exists>F'. E' = cCallR t l F' \<and> E \<approx> F'" 
      using prems(1)
    proof (cases E', auto)
      assume "Call t l (E\<up>x) = x" hence "x = (cCallR t l E)\<up>x" by auto
      from fixpoint_context[OF this] show False by simp
    next
      assume "E\<up>x = E\<up>y" 
      from eq_context[OF this] `x \<noteq> y` show False by simp
    qed
  qed
next
  case (cUpdL E l t) show ?case
  proof (simp, intro strip, elim conjE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "x \<noteq> y"
           "Upd (E\<up>x) l t = E'\<up>x"
           "Upd (E\<up>y) l t = E'\<up>y"
    thus "\<exists>F'. E' = cUpdL F' l t \<and> E \<approx> F'" 
      using prems(1)
    proof (cases E', auto)
      assume "Upd (E\<up>x) l t = x" hence "x = (cUpdL E l t)\<up>x" by auto
      from fixpoint_context[OF this] show False by simp
    next
      assume "E\<up>y = E\<up>x"
      from eq_context[OF this] `x \<noteq> y` show False by simp
    qed
  qed
next
  case (cUpdR t l E) show ?case
  proof (simp, intro strip, elim conjE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "x \<noteq> y"
           "Upd t l (E\<up>x) = E'\<up>x"
           "Upd t l (E\<up>y) = E'\<up>y"
    thus "\<exists>F'. E' = cUpdR t l F' \<and> E \<approx> F'" 
      using prems(1)
    proof (cases E', auto)
      assume "Upd t l (E\<up>x) = x" hence "x = (cUpdR t l E)\<up>x" by auto
      from fixpoint_context[OF this] show False by simp
    next
      assume "E\<up>x = E\<up>y" 
      from eq_context[OF this] `x \<noteq> y` show False by simp
    qed
  qed
next
  case (cActive E) show ?case
  proof (simp, intro strip, elim conjE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "x \<noteq> y"
           "Active (E\<up>x) = E'\<up>x"
           "Active (E\<up>y) = E'\<up>y"
    thus "\<exists>F'. E' = cActive F' \<and> E \<approx> F'" 
      using prems(1)
    proof (cases E', auto)
      assume "Active (E\<up>x) = x" hence "x = (cActive E)\<up>x" by auto
      from fixpoint_context[OF this] show False by simp
    qed
  qed
qed

(* useful consequence of the equality of 2 filled contexts *)
lemma contexts_and_sterm[rule_format]: 
  "\<forall>x y (E'::general_context). (E::general_context)\<up>x = E'\<up>y 
    \<longrightarrow> ((\<exists>(E1::general_context). x = E1\<up>y) 
       \<or> (\<exists>(E1::general_context). E1\<up>x = y)
       \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y))"
proof (induct E)
  case cHole thus ?case
  proof (simp, intro allI disjI1)
    fix y :: sterm and E' :: general_context
    show "\<exists>(E1::general_context). E'\<up>y = E1\<up>y" by (rule_tac x = E' in exI, simp)
  qed
next
  case (cObj f T E) thus ?case
  proof (simp, intro allI impI, elim exE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "\<forall>x y. (\<exists>(E'::general_context). E\<up>x = E'\<up>y)
             \<longrightarrow> (\<exists>(E1::general_context). x = E1\<up>y) 
               \<or> (\<exists>(E1::general_context). E1\<up>x = y)
               \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
           "Obj (FLmap f(FLlabel f \<mapsto> E\<up>x)) T = E'\<up>y"
    thus "(\<exists>(E1::general_context). x = E1\<up>y)
        \<or> (\<exists>(E1::general_context). E1\<up>x = y)
        \<or> (\<forall>z. \<exists>(E2::general_context). Obj (FLmap f(FLlabel f \<mapsto> E\<up>z)) T = E2\<up>y)"
    proof (cases E', simp_all add: prems(2) del: sterm.inject(3))
      case cHole with `Obj (FLmap f(FLlabel f \<mapsto> E\<up>x)) T = E'\<up>y`
      have "Obj (FLmap f(FLlabel f \<mapsto> E\<up>x)) T = y" by simp
      hence "\<exists>(E1::general_context). E1\<up>x = y"
	by (rule_tac x = "cObj f T E" in exI, simp)
      thus ?thesis by simp
    next
      case (cObj fa Ta Ea) with `Obj (FLmap f(FLlabel f \<mapsto> E\<up>x)) T = E'\<up>y`
      have ffa: "FLmap f(FLlabel f \<mapsto> E\<up>x) = FLmap fa(FLlabel fa \<mapsto> Ea\<up>y)"
	by simp
      thus ?thesis
      proof (cases "FLlabel f = FLlabel fa")
	case True with ffa have "E\<up>x = Ea\<up>y" 
	  using gc_obj_access_filled[of f E x] 
	        gc_obj_access_filled[of fa Ea y]
	  by simp
	with prems(1)
	have "(\<exists>(E1::general_context). x = E1\<up>y) 
            \<or> (\<exists>(E1::general_context). E1\<up>x = y) 
	    \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)" by blast
	thus ?thesis
	proof (elim disjE, simp_all, intro disjI2 allI)
	  fix z :: sterm
	  assume "\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y"
	  then obtain E2 where "E\<up>z = (E2::general_context)\<up>y" by auto
	  thus "\<exists>(E2::general_context). Obj (FLmap f(FLlabel f \<mapsto> E\<up>z)) T = E2\<up>y"
	    by (rule_tac x = "cObj f T E2" in exI, simp)
	qed
      next
	case False thus ?thesis
	proof (intro disjI2 allI)
	  fix z :: sterm
	  {
	    fix l :: Label
	    note fun_upd_other[OF not_sym[OF `FLlabel f \<noteq> FLlabel fa`]]
	    from this[of "FLmap f" "Some (E\<up>x)"] ffa `FLlabel f \<noteq> FLlabel fa`
	    have "(FLmap f(FLlabel f \<mapsto> E\<up>z)) l 
	          = (FLmap f(FLlabel f \<mapsto> E\<up>z, FLlabel fa \<mapsto> Ea\<up>y)) l"
	      by (cases "l = FLlabel fa", auto)
	  }
	  moreover
	  note arg_cong[OF ffa]
	  from this[of "\<lambda>x. x (FLlabel fa)"]
	       gc_obj_access_filled[of fa Ea y]
	  have "(FLmap f(FLlabel f \<mapsto> E\<up>z), FLlabel fa) \<in> FmapLabel"
	    unfolding FmapLabel_def by auto
	  note Abs_FmapLabel_inverse[OF this]
	  ultimately
	  show "\<exists>(E2::general_context). Obj (FLmap f(FLlabel f \<mapsto> E\<up>z)) T = E2\<up>y"
	    by (rule_tac x = "cObj (Abs_FmapLabel(FLmap f(FLlabel f \<mapsto> E\<up>z),FLlabel fa)) T Ea" 
	         in exI, simp add: ext FLmap_def FLlabel_def)
	qed
      qed
    qed
  qed
next
  case (cCallL E l t) thus ?case
  proof (simp, intro allI impI, elim exE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "\<forall>x y. (\<exists>(E'::general_context). E\<up>x = E'\<up>y)
             \<longrightarrow> (\<exists>(E1::general_context). x = E1\<up>y) 
               \<or> (\<exists>(E1::general_context). E1\<up>x = y)
               \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
           "Call (E\<up>x) l t = E'\<up>y"
    thus "(\<exists>(E1::general_context). x = E1\<up>y)
        \<or> (\<exists>(E1::general_context). E1\<up>x = y)
        \<or> (\<forall>z. \<exists>(E2::general_context). Call (E\<up>z) l t = E2\<up>y)"
    proof (cases E', simp_all add: prems(2) del: sterm.inject(4))
      case cHole moreover assume "Call (E\<up>x) l t = y"
      ultimately have "\<exists>(E1::general_context). E1\<up>x = y"
	by (rule_tac x = "cCallL E l t" in exI, simp)
      thus ?thesis by simp
    next
      case (cCallR ta la Ea) with `Call (E\<up>x) l t = E'\<up>y` show ?thesis
	by (intro disjI2 allI, rule_tac x = "cCallR (E\<up>z) l Ea" in exI, simp)
    next
      case (cCallL Ea la ta) with prems(2) `Call (E\<up>x) l t = E'\<up>y`
      have "(\<exists>(E1::general_context). x = E1\<up>y) 
	  \<or> (\<exists>(E1::general_context). E1\<up>x = y) 
	  \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	by auto
      thus ?thesis proof (auto)
	fix z :: sterm
	assume "(\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	then obtain E2 :: general_context where "E\<up>z = E2\<up>y" by auto
	thus "\<exists>(E2::general_context). Call (E\<up>z) l t = E2\<up>y"
	  by (rule_tac x = "cCallL E2 l t" in exI, simp)
      qed
    qed
  qed
next
  case (cCallR t l E) thus ?case
  proof (simp, intro allI impI, elim exE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "\<forall>x y. (\<exists>(E'::general_context). E\<up>x = E'\<up>y)
             \<longrightarrow> (\<exists>(E1::general_context). x = E1\<up>y) 
               \<or> (\<exists>(E1::general_context). E1\<up>x = y)
               \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
           "Call t l (E\<up>x) = E'\<up>y"
    thus "(\<exists>(E1::general_context). x = E1\<up>y)
        \<or> (\<exists>(E1::general_context). E1\<up>x = y)
        \<or> (\<forall>z. \<exists>(E2::general_context). Call t l (E\<up>z) = E2\<up>y)"
    proof (cases E', simp_all add: prems(2) del: sterm.inject(4))
      case cHole moreover assume "Call t l (E\<up>x) = y"
      ultimately have "\<exists>(E1::general_context). E1\<up>x = y"
	by (rule_tac x = "cCallR t l E" in exI, simp)
      thus ?thesis by simp
    next
      case (cCallL Ea la ta) with `Call t l (E\<up>x) = E'\<up>y` show ?thesis
	by (intro disjI2 allI, rule_tac x = "cCallL Ea l (E\<up>z)" in exI, simp)
    next
      case (cCallR ta la Ea) with prems(2) `Call t l (E\<up>x) = E'\<up>y`
      have "(\<exists>(E1::general_context). x = E1\<up>y) 
	  \<or> (\<exists>(E1::general_context). E1\<up>x = y) 
	  \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	by auto
      thus ?thesis proof (auto)
	fix z :: sterm
	assume "(\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	then obtain E2 :: general_context where "E\<up>z = E2\<up>y" by auto
	thus "\<exists>(E2::general_context). Call t l (E\<up>z) = E2\<up>y"
	  by (rule_tac x = "cCallR t l E2" in exI, simp)
      qed
    qed
  qed
next
  case (cUpdL E l t) thus ?case
  proof (simp, intro allI impI, elim exE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "\<forall>x y. (\<exists>(E'::general_context). E\<up>x = E'\<up>y)
             \<longrightarrow> (\<exists>(E1::general_context). x = E1\<up>y) 
               \<or> (\<exists>(E1::general_context). E1\<up>x = y)
               \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
           "Upd (E\<up>x) l t = E'\<up>y"
    thus "(\<exists>(E1::general_context). x = E1\<up>y)
        \<or> (\<exists>(E1::general_context). E1\<up>x = y)
        \<or> (\<forall>z. \<exists>(E2::general_context). Upd (E\<up>z) l t = E2\<up>y)"
    proof (cases E', simp_all add: prems(2) del: sterm.inject(5))
      case cHole moreover assume "Upd (E\<up>x) l t = y"
      ultimately have "\<exists>(E1::general_context). E1\<up>x = y"
	by (rule_tac x = "cUpdL E l t" in exI, simp)
      thus ?thesis by simp
    next
      case (cUpdR ta la Ea) with `Upd (E\<up>x) l t = E'\<up>y` show ?thesis
	by (intro disjI2 allI, rule_tac x = "cUpdR (E\<up>z) l Ea" in exI, simp)
    next
      case (cUpdL Ea la ta) with prems(2) `Upd (E\<up>x) l t = E'\<up>y`
      have "(\<exists>(E1::general_context). x = E1\<up>y) 
	  \<or> (\<exists>(E1::general_context). E1\<up>x = y) 
	  \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	by auto
      thus ?thesis proof (auto)
	fix z :: sterm
	assume "(\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	then obtain E2 :: general_context where "E\<up>z = E2\<up>y" by auto
	thus "\<exists>(E2::general_context). Upd (E\<up>z) l t = E2\<up>y"
	  by (rule_tac x = "cUpdL E2 l t" in exI, simp)
      qed
    qed
  qed
next
  case (cUpdR t l E) thus ?case
  proof (simp, intro allI impI, elim exE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "\<forall>x y. (\<exists>(E'::general_context). E\<up>x = E'\<up>y)
             \<longrightarrow> (\<exists>(E1::general_context). x = E1\<up>y) 
               \<or> (\<exists>(E1::general_context). E1\<up>x = y)
               \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
           "Upd t l (E\<up>x) = E'\<up>y"
    thus "(\<exists>(E1::general_context). x = E1\<up>y)
        \<or> (\<exists>(E1::general_context). E1\<up>x = y)
        \<or> (\<forall>z. \<exists>(E2::general_context). Upd t l (E\<up>z) = E2\<up>y)"
    proof (cases E', simp_all add: prems(2) del: sterm.inject(5))
      case cHole moreover assume "Upd t l (E\<up>x) = y"
      ultimately have "\<exists>(E1::general_context). E1\<up>x = y"
	by (rule_tac x = "cUpdR t l E" in exI, simp)
      thus ?thesis by simp
    next
      case (cUpdL Ea la ta) with `Upd t l (E\<up>x) = E'\<up>y` show ?thesis
	by (intro disjI2 allI, rule_tac x = "cUpdL Ea l (E\<up>z)" in exI, simp)
    next
      case (cUpdR ta la Ea) with prems(2) `Upd t l (E\<up>x) = E'\<up>y`
      have "(\<exists>(E1::general_context). x = E1\<up>y) 
	  \<or> (\<exists>(E1::general_context). E1\<up>x = y) 
	  \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	by auto
      thus ?thesis proof (auto)
	fix z :: sterm
	assume "(\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	then obtain E2 :: general_context where "E\<up>z = E2\<up>y" by auto
	thus "\<exists>(E2::general_context). Upd t l (E\<up>z) = E2\<up>y"
	  by (rule_tac x = "cUpdR t l E2" in exI, simp)
      qed
    qed
  qed
next
  case (cActive E) thus ?case
  proof (simp, intro allI impI, elim exE)
    fix x :: sterm and y :: sterm and E' :: general_context
    assume "\<forall>x y. (\<exists>(E'::general_context). E\<up>x = E'\<up>y)
             \<longrightarrow> (\<exists>(E1::general_context). x = E1\<up>y) 
               \<or> (\<exists>(E1::general_context). E1\<up>x = y)
               \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
           "Active (E\<up>x) = E'\<up>y"
    thus "(\<exists>(E1::general_context). x = E1\<up>y)
        \<or> (\<exists>(E1::general_context). E1\<up>x = y)
        \<or> (\<forall>z. \<exists>(E2::general_context). Active (E\<up>z) = E2\<up>y)"
    proof (cases E', simp_all add: prems(2))
      case cHole moreover assume "Active (E\<up>x) = y"
      ultimately have "\<exists>(E1::general_context). E1\<up>x = y"
	by (rule_tac x = "cActive E" in exI, simp)
      thus ?thesis by simp
    next
      case (cActive Ea) with prems(2) `Active (E\<up>x) = E'\<up>y`
      have "(\<exists>(E1::general_context). x = E1\<up>y) 
	  \<or> (\<exists>(E1::general_context). E1\<up>x = y) 
	  \<or> (\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	by auto
      thus ?thesis proof (auto) 
	fix z :: sterm
	assume "(\<forall>z. \<exists>(E2::general_context). E\<up>z = E2\<up>y)"
	then obtain E2 :: general_context where "E\<up>z = E2\<up>y" by auto
	thus "\<exists>(E2::general_context). Active (E\<up>z) = E2\<up>y"
	  by (rule_tac x = "cActive E2" in exI, simp)
      qed
    qed
  qed
qed

(* FORMATTED : lemma contexts_and_sterm:
        ?E \<up> ?x = ?E' \<up> ?y \<Longrightarrow>
        (\<exists>E1. ?x = E1 \<up> ?y) \<or> (\<exists>E1. E1 \<up> ?x = ?y) \<or> (\<forall>z. \<exists>E2. ?E \<up> z = E2 \<up> ?y)   *)

(* if filling gives a free variable *)
lemma context_fvar[rule_format]:
  "\<forall>x y. ((E::general_context)\<up>x) = Fvar y
    \<longrightarrow> E = cHole \<and> x = Fvar y"
  by (case_tac E, auto)

(* if filling gives a bound variable *)
lemma context_bvar[rule_format]:
  "\<forall>x b. ((E::general_context)\<up>x) = Bvar b
    \<longrightarrow> E = cHole \<and> x = Bvar b"
  by (case_tac E, auto)

(* if filling gives a Call *)
lemma context_call[rule_format]: 
  "\<forall>x li d b. (E::general_context)\<up>x = Call d li b
    \<longrightarrow> (E = cHole \<or> (\<exists>(E'::general_context). d = E'\<up>x))
      \<or> (\<exists>(E'::general_context). b = E'\<up>x)"
  by (case_tac E, auto)

lemma ref_context_call: 
  fixes E :: general_context
  assumes "isref r" "E\<up>r = Call t l t'"
  shows "(\<exists>E'::general_context. t = E'\<up>r) \<or> (\<exists>E'::general_context. t' = E'\<up>r)"
proof -
  from `isref r` context_call[OF assms(2)] `E\<up>r = Call t l t'`
  show ?thesis by (cases rule: isref.cases, auto)
qed

lemma context_call_inverse:
  "(\<exists>E::general_context. t1 = E\<up>t) \<or> (\<exists>E::general_context. t2 = E\<up>t) 
  \<Longrightarrow> \<exists>E'::general_context. Call t1 l t2 = E'\<up>t"
proof (elim disjE exE)
  fix E::general_context 
  def ctxt \<equiv> "cCallL E l t2"
  assume "t1 = E\<up>t" hence "Call t1 l t2 = ctxt\<up>t"
    unfolding ctxt_def by simp
  thus ?thesis by blast
next
  fix E::general_context 
  def ctxt \<equiv> "cCallR t1 l E"
  assume "t2 = E\<up>t" hence "Call t1 l t2 = ctxt\<up>t"
    unfolding ctxt_def by simp
  thus ?thesis by blast
qed

(* if filling results in Upd *)
lemma context_upd[rule_format]: 
  "\<forall>x li a b. (E::general_context)\<up>x = Upd a li b
    \<longrightarrow> (E = cHole \<or> (\<exists>(E'::general_context). a = E'\<up>x))
      \<or> (\<exists>(E'::general_context). b = E'\<up>x)"
  by (case_tac E, auto)

lemma ref_context_upd: 
  fixes E :: general_context
  assumes "isref r" "E\<up>r = Upd t l t'"
  shows "(\<exists>E'::general_context. t = E'\<up>r) \<or> (\<exists>E'::general_context. t' = E'\<up>r)"
proof -
  from `isref r` context_upd[OF assms(2)] `E\<up>r = Upd t l t'`
  show ?thesis by (cases rule: isref.cases, auto)
qed

lemma context_upd_inverse:
  "(\<exists>E::general_context. t1 = E\<up>t) \<or> (\<exists>E::general_context. t2 = E\<up>t) 
  \<Longrightarrow> \<exists>E'::general_context. Upd t1 l t2 = E'\<up>t"
proof (elim disjE exE)
  fix E::general_context 
  def ctxt \<equiv> "cUpdL E l t2"
  assume "t1 = E\<up>t" hence "Upd t1 l t2 = ctxt\<up>t"
    unfolding ctxt_def by simp
  thus ?thesis by blast
next
  fix E::general_context 
  def ctxt \<equiv> "cUpdR t1 l E"
  assume "t2 = E\<up>t" hence "Upd t1 l t2 = ctxt\<up>t"
    unfolding ctxt_def by simp
  thus ?thesis by blast
qed

lemma context_obj[rule_format]: 
  "\<forall>x f T. ((E::general_context)\<up>x) = Obj f T 
    \<longrightarrow> E = cHole \<or> (\<exists>l (E'::general_context). (f l) = Some (E'\<up>x))"
  by (case_tac E, auto)

lemma ref_context_obj: 
  fixes E :: general_context
  assumes "isref r" "E\<up>r = Obj f T"
  shows "\<exists>l E'::general_context. f l = Some (E'\<up>r)"
proof -
  from `isref r` context_obj[OF assms(2)] `E\<up>r = Obj f T`
  show ?thesis by (cases rule: isref.cases, auto)
qed

lemma context_obj_inverse: 
  assumes "\<exists>l E::general_context. f l = Some (E\<up>t)"
  shows "\<exists>E'::general_context. Obj f T = E'\<up>t"
proof -
  from assms(1) obtain l and E::general_context
    where "f l = Some (E\<up>t)" by auto
  def ctxt \<equiv> "cObj (Abs_FmapLabel (f,l)) T E"
  from `f l = Some (E\<up>t)` have "(f,l) \<in> FmapLabel"
    by (auto simp: FmapLabel_def)
  from `f l = Some (E\<up>t)` Abs_FmapLabel_inverse[OF this]
  have "Obj f T = ctxt\<up>t" unfolding ctxt_def
    by (auto simp: FLmap_def FLlabel_def map_upd_triv)
  thus ?thesis by blast
qed

lemma context_active[rule_format]:
  "\<forall>x a. ((E::general_context)\<up>x) = Active a
    \<longrightarrow> E = cHole \<or> (\<exists>(E'::general_context). a = E'\<up>x)"
  by (case_tac E, auto)

lemma ref_context_active: 
  fixes E :: general_context
  assumes "isref r" "E\<up>r = Active a"
  shows "\<exists>E'::general_context. a = E'\<up>r"
proof -
  from `isref r` context_active[OF assms(2)] `E\<up>r = Active a`
  show ?thesis by (cases rule: isref.cases, auto)
qed

lemma context_active_inverse: 
  assumes "\<exists>E::general_context. a = E\<up>t"
  shows "\<exists>E'::general_context. Active a = E'\<up>t"
proof -
  from assms(1) obtain E::general_context where "a = E\<up>t"
    by auto
  def ctxt \<equiv> "cActive E"
  from `a = E\<up>t` have "Active a = ctxt\<up>t"
    unfolding ctxt_def by simp
  thus ?thesis by blast
qed

section {* lemmas concerning references *}
lemma context_actrefs: "ActRef A = E \<up> b \<Longrightarrow> E = cHole \<and> b = ActRef A"
  by (case_tac E, auto)

lemma context_futrefs: "FutRef f = E \<up> b \<Longrightarrow> E = cHole \<and> b = FutRef f"
  by (case_tac E, auto)

lemma diff_actrefs: "A \<noteq> B \<Longrightarrow> \<forall>E::general_context. ActRef A \<noteq> E \<up>ActRef B"
  by (intro allI, case_tac E, auto)

lemma diff_futrefs: "fk \<noteq> fl \<Longrightarrow> \<forall>E::general_context. FutRef fk \<noteq> E\<up>FutRef fl"
  by (rule allI, case_tac E, auto)

lemma diff_futactrefs: "\<forall>E::general_context. ActRef G \<noteq> E\<up>FutRef fl"
  by (rule allI, case_tac E, auto)

lemma diff_actfutrefs: "\<forall>E::general_context. FutRef fk \<noteq> E\<up>ActRef G"
  by (rule allI, case_tac E, auto)

lemma context_inc_Ref:
fixes E :: general_context and t :: sterm and E' :: general_context and r :: sterm
assumes "isref r" "E\<up>t = E'\<up>r"
shows "(\<exists>(E1::general_context). t = E1\<up>r)
      \<or>(\<forall>b.\<exists>(E2::general_context). E\<up>b = E2\<up>r)"
proof -
  from contexts_and_sterm[OF assms(2)] show ?thesis
  proof (elim disjE, simp_all, elim exE)
    fix Et :: general_context assume "Et\<up>t = r"
    with `isref r` context_actrefs[of _ Et t] context_futrefs[of _ Et t]
    have "\<exists>Ea::general_context. t = Ea\<up>r"
      by (rule_tac x = cHole in exI, 
	cases rule: isref.cases, auto)
    thus ?thesis by simp
  qed
qed

lemma context_inc_Act: "(E::general_context)\<up>t = (E'::general_context)\<up>ActRef A 
  \<Longrightarrow> (\<exists>(E1::general_context). t = E1\<up>ActRef A)\<or>(\<forall>b.\<exists>(E2::general_context). E\<up>b = E2\<up>ActRef A)"
  by (auto simp: context_inc_Ref[of "ActRef A" E t E'])

lemma context_inc_Fut: "(E::general_context)\<up>t = (E'::general_context)\<up>FutRef f 
  \<Longrightarrow> (\<exists>(E1::general_context). t = E1\<up>FutRef f)\<or>(\<forall>b.\<exists>(E2::general_context). E\<up>b = E2\<up>FutRef f)"
  by (auto simp: context_inc_Ref[of "FutRef f" E t E'])

section {* econtext lemmas *}
lemma eq_econtext: "(E::econtext)\<up>x = E\<up>y \<Longrightarrow> x = y"
proof -
  fix E :: econtext and x :: sterm and y :: sterm
  assume "E\<up>x = E\<up>y" with ectxt2gctxt_and_fill
  have "(ectxt2gctxt E)\<up>x = (ectxt2gctxt E)\<up>y" by simp
  from eq_context[OF this] show "x = y" by assumption
qed

subsection {* Context opening *}
primrec
  copen        :: "[nat, sterm, sterm, general_context] \<Rightarrow> general_context" 
                                      ("{_ \<rightarrow>\<^sub>c [_,_]} _" [0, 0, 0, 300] 300)
where
   copen_Hole:  "{k \<rightarrow>\<^sub>c [s,p]} cHole = cHole"
 | copen_Obj:   "{k \<rightarrow>\<^sub>c [s,p]}(cObj f T E) =
                 cObj (Abs_FmapLabel ((\<lambda>l. sopen_option (Suc k) s p (FLmap f l)), FLlabel f)) T 
                      ({(Suc k) \<rightarrow>\<^sub>c [s,p]} E)"
 | copen_CallL: "{k \<rightarrow>\<^sub>c [s,p]}(cCallL E l t) = cCallL ({k \<rightarrow>\<^sub>c [s,p]} E) l ({k \<rightarrow> [s,p]} t)"
 | copen_CallR: "{k \<rightarrow>\<^sub>c [s,p]}(cCallR t l E) = cCallR ({k \<rightarrow> [s,p]} t) l ({k \<rightarrow>\<^sub>c [s,p]} E)"
 | copen_UpdL:  "{k \<rightarrow>\<^sub>c [s,p]}(cUpdL E l t) = cUpdL ({k \<rightarrow>\<^sub>c [s,p]} E) l ({(Suc k) \<rightarrow> [s,p]} t)"
 | copen_UpdR:  "{k \<rightarrow>\<^sub>c [s,p]}(cUpdR t l E) = cUpdR ({k \<rightarrow> [s,p]} t) l ({(Suc k) \<rightarrow>\<^sub>c [s,p]} E)"
 | copen_Act:   "{k \<rightarrow>\<^sub>c [s,p]}(cActive E) = cActive ({k \<rightarrow>\<^sub>c [s,p]} E)"

lemma sopen_and_copen[rule_format]: "\<forall>n s p t. lc t \<longrightarrow> {n \<rightarrow> [s,p]} (E\<up>t) = {n \<rightarrow>\<^sub>c [s,p]} E\<up>t"
  apply (induct E, auto)
  apply (rule Ltake_eq_all, auto)
  apply (unfold Ltake_eq_def, auto)
done

subsection {* Context closing *}
primrec
  cclose        :: "[nat, fVariable, fVariable, general_context] \<Rightarrow> general_context" 
                                      ("{_ \<leftarrow>\<^sub>c [_,_]} _" [0, 0, 0, 300] 300)
where
   cclose_Hole:  "{k \<leftarrow>\<^sub>c [s,p]} cHole = cHole"
 | cclose_Obj:   "{k \<leftarrow>\<^sub>c [s,p]}(cObj f T E) =
                 cObj (Abs_FmapLabel ((\<lambda>l. sclose_option (Suc k) s p (FLmap f l)), FLlabel f)) T 
                      ({(Suc k) \<leftarrow>\<^sub>c [s,p]} E)"
 | cclose_CallL: "{k \<leftarrow>\<^sub>c [s,p]}(cCallL E l t) = cCallL ({k \<leftarrow>\<^sub>c [s,p]} E) l ({k \<leftarrow> [s,p]} t)"
 | cclose_CallR: "{k \<leftarrow>\<^sub>c [s,p]}(cCallR t l E) = cCallR ({k \<leftarrow> [s,p]} t) l ({k \<leftarrow>\<^sub>c [s,p]} E)"
 | cclose_UpdL:  "{k \<leftarrow>\<^sub>c [s,p]}(cUpdL E l t) = cUpdL ({k \<leftarrow>\<^sub>c [s,p]} E) l ({(Suc k) \<leftarrow> [s,p]} t)"
 | cclose_UpdR:  "{k \<leftarrow>\<^sub>c [s,p]}(cUpdR t l E) = cUpdR ({k \<leftarrow> [s,p]} t) l ({(Suc k) \<leftarrow>\<^sub>c [s,p]} E)"
 | cclose_Act:   "{k \<leftarrow>\<^sub>c [s,p]}(cActive E) = cActive ({k \<leftarrow>\<^sub>c [s,p]} E)"

lemma sclose_and_cclose[rule_format]: 
  "\<forall>n s p t. s \<notin> FV t \<longrightarrow> p \<notin> FV t \<longrightarrow> s \<noteq> p \<longrightarrow> lc t 
    \<longrightarrow> {n \<leftarrow> [s,p]} (E\<up>t) = {n \<leftarrow>\<^sub>c [s,p]} E\<up>t"
  apply (induct E, auto)
  apply (rule Ltake_eq_all, auto)
  apply (unfold Ltake_eq_def, auto)
done

subsection {* Free variables on contexts *}
primrec
 FV_ctxt  :: "general_context \<Rightarrow> fVariable set"
where
   FV_ctxt_Hole  : "FV_ctxt cHole = {}"
 | FV_ctxt_Obj   : "FV_ctxt (cObj f T E) 
                    = (\<Union>l\<in>dom (FLmap f). FVoption (FLmap f l)) \<union> FV_ctxt E"
 | FV_ctxt_CallL : "FV_ctxt (cCallL E l t) = FV_ctxt E \<union> FV t"
 | FV_ctxt_CallR : "FV_ctxt (cCallR t l E) = FV t \<union> FV_ctxt E"
 | FV_ctxt_UpdL  : "FV_ctxt (cUpdL E l t) = FV_ctxt E \<union> FV t"
 | FV_ctxt_UpdR  : "FV_ctxt (cUpdR t l E) = FV t \<union> FV_ctxt E"
 | FV_ctxt_Act   : "FV_ctxt (cActive E) = FV_ctxt E"

lemma finite_FV_ctxt[simp]: "finite (FV_ctxt E)"
  apply (induct_tac E, auto)
  apply (subgoal_tac "finite (dom (FLmap FmapLabel))", simp add: finite_FV_FVoption)
  apply (simp add: finite_Label finite_dom_fmap)
done

lemma FV_ctxt_and_Fill:
  "FV ((E::general_context)\<up>t) \<subseteq> FV_ctxt E \<union> FV t"
proof (induct E)
  case cHole thus ?case by auto
next
  case cCallL thus ?case by auto
next
  case cCallR thus ?case by auto
next
  case cActive thus ?case by auto
next
  case cUpdL thus ?case by auto
next
  case cUpdR thus ?case by auto
next
  case (cObj f T E) note pred = this
  {
    fix x l
    assume 
      "l \<in> dom (FLmap f)" and "x \<notin> FV_ctxt E" and "x \<notin> FV t" and
      x_in: "x \<in> FVoption ((FLmap f(FLlabel f \<mapsto> E\<up>t)) l)"
    hence "\<exists>l\<in>dom (FLmap f). x \<in> FVoption (FLmap f l)"
    proof (cases "l = (FLlabel f)")
      case True with gc_obj_access_filled[of f E t] x_in
      have "x \<in> FV (E\<up>t)" by auto
      with `x \<notin> FV_ctxt E` `x \<notin> FV t` pred
      show ?thesis by auto
    next
      case False with `l \<in> dom (FLmap f)` x_in show ?thesis by auto
    qed
  } note in_FVoption = this

  from pred show ?case by (auto, intro in_FVoption, auto)
qed

lemma FV_ctxt_and_Fill':
  "FV t \<subseteq> FV ((E::general_context)\<up>t)"
  by (induct E, auto)

subsection {* local closure on general contexts *}
inductive lc_ctxt :: "general_context \<Rightarrow> bool"
where
    lc_ctxt_Hole[simp, intro!] : "lc_ctxt cHole"
  | lc_ctxt_Obj[simp, intro!]  : "\<lbrakk> finite L; \<forall>l\<in>dom (FLmap f).\<forall>s p. s \<notin> L \<and> p \<notin> L \<and> s \<noteq> p 
                                                    \<longrightarrow> lc (the(FLmap f l)\<^bsup>[Fvar s, Fvar p]\<^esup>);
                                    \<forall>s p. s \<notin> L \<and> p \<notin> L \<and> s \<noteq> p
                                     \<longrightarrow> lc_ctxt ({0 \<rightarrow>\<^sub>c [Fvar s, Fvar p]} E) \<rbrakk>
                                 \<Longrightarrow> lc_ctxt (cObj f T E)"
  | lc_ctxt_CallL[simp, intro!]: "\<lbrakk> lc_ctxt E; lc a \<rbrakk> \<Longrightarrow> lc_ctxt (cCallL E l a)"
  | lc_ctxt_CallR[simp, intro!]: "\<lbrakk> lc_ctxt E; lc t \<rbrakk> \<Longrightarrow> lc_ctxt (cCallR t l E)"
  | lc_ctxt_UpdL[simp, intro!] : "\<lbrakk> lc_ctxt E; finite L; \<forall>s p. s \<notin> L \<and> p \<notin> L \<and> s \<noteq> p 
                                                            \<longrightarrow> lc (u\<^bsup>[Fvar s, Fvar p]\<^esup>) \<rbrakk>
                                 \<Longrightarrow> lc_ctxt (cUpdL E l u)"
  | lc_ctxt_UpdR[simp, intro!] : "\<lbrakk> lc t; finite L; \<forall>s p. s \<notin> L \<and> p \<notin> L \<and> s \<noteq> p 
                                                      \<longrightarrow> lc_ctxt ({0 \<rightarrow>\<^sub>c [Fvar s, Fvar p]} E) \<rbrakk>
                                 \<Longrightarrow> lc_ctxt (cUpdR t l E)"
  | lc_ctxt_Act[simp, intro!]  : "lc_ctxt E \<Longrightarrow> lc_ctxt (cActive E)"

lemma lc_ctxt_and_lc[rule_format]: "lc_ctxt E \<Longrightarrow> \<forall>t. lc t \<longrightarrow> lc (E\<up>t)"
proof -
  {
    fix L :: "fVariable \<Rightarrow> bool" and E :: general_context and
        t :: sterm and s :: fVariable and p :: fVariable
    assume "\<forall>s p. s \<notin> L \<and> p \<notin> L \<and> s \<noteq> p 
             \<longrightarrow> lc_ctxt ({0 \<rightarrow>\<^sub>c [Fvar s,Fvar p]} E) 
               \<and> (\<forall>t. lc t \<longrightarrow> lc ({0 \<rightarrow>\<^sub>c [Fvar s,Fvar p]} E\<up>t))"
           "lc t" "s \<notin> L" "p \<notin> L" "s \<noteq> p"
    hence "lc ({0 \<rightarrow>\<^sub>c [Fvar s,Fvar p]} E\<up>t)" by blast
    with sopen_and_copen[OF `lc t`] 
    have "lc ((E\<up>t)\<^bsup>[Fvar s,Fvar p]\<^esup>)" by (simp add: openz_def)
  }
  note lem = this

(* induction *)
  show "lc_ctxt E \<Longrightarrow> \<forall>t. lc t \<longrightarrow> lc (E\<up>t)"
  proof (induct E rule: lc_ctxt.induct, auto)
    case (lc_ctxt_Obj L f E T)
    moreover
    fix l :: Label and t :: sterm and s :: fVariable and p :: fVariable and
        y :: sterm
    assume "lc t"  and
   Some_y: "(if l = FLlabel f then Some (E \<up> t) else FLmap f l) = Some y" and
           "s \<notin> L" "p \<notin> L" "s \<noteq> p"
    ultimately show "lc (y\<^bsup>[Fvar s,Fvar p]\<^esup>)"
    proof (cases "l = FLlabel f")
      case False with Some_y 
      have "l \<in> dom (FLmap f)" by force 
      with `s \<notin> L` `p \<notin> L` `s \<noteq> p` prems(2)
      have "lc (the (FLmap f l)\<^bsup>[Fvar s,Fvar p]\<^esup>)" by blast
      with Some_y `l \<noteq> FLlabel f` show "lc (y\<^bsup>[Fvar s,Fvar p]\<^esup>)" by auto
    next
      case True with lem[OF prems(3) `lc t` `s \<notin> L` `p \<notin> L` `s \<noteq> p`] Some_y
      show "lc (y\<^bsup>[Fvar s,Fvar p]\<^esup>)" by auto
    qed
  next
    case (lc_ctxt_UpdR t L E l)
    moreover
    fix t' :: sterm and s :: fVariable and p :: fVariable
    assume "lc t'" "s \<notin> L" "p \<notin> L" "s \<noteq> p"
    from lem[OF prems(3) this]
    show "lc (E\<up>t'\<^bsup>[Fvar s,Fvar p]\<^esup>)" by assumption
  qed
qed

lemma beta_lc_ctxt[simp]: "lc_ctxt E \<Longrightarrow> \<forall>t t'. t \<rightarrow>\<^sub>\<beta> t' \<longrightarrow> E\<up>t \<rightarrow>\<^sub>\<beta> E\<up>t'"
proof -
  {
    fix L :: "fVariable \<Rightarrow> bool" and E :: general_context and
        t :: sterm and t' :: sterm
    assume "\<forall>s p. s \<notin> L \<and> p \<notin> L \<and> s \<noteq> p 
             \<longrightarrow> lc_ctxt ({0 \<rightarrow>\<^sub>c [Fvar s,Fvar p]} E) 
	      \<and> (\<forall>t t'. t \<rightarrow>\<^sub>\<beta> t' 
	          \<longrightarrow> {0 \<rightarrow>\<^sub>c [Fvar s,Fvar p]} E\<up>t \<rightarrow>\<^sub>\<beta> {0 \<rightarrow>\<^sub>c [Fvar s,Fvar p]} E\<up>t')"
           "t \<rightarrow>\<^sub>\<beta> t'"
    hence "\<forall>s p. s \<notin> L \<union> FV (E\<up>t') \<and> p \<notin> L \<union> FV (E\<up>t') \<and> s \<noteq> p 
            \<longrightarrow> (\<exists>t''. E\<up>t\<^bsup>[Fvar s,Fvar p]\<^esup> \<rightarrow>\<^sub>\<beta> t'' \<and> E\<up>t' = \<sigma>[s,p] t'')"
    proof (intro strip, simp, elim conjE)
      fix s :: fVariable and p :: fVariable
      assume "s \<notin> L" "s \<notin> FV (E\<up>t')" "p \<notin> L" "p \<notin> FV (E\<up>t')" "s \<noteq> p"
      with prems(1) `t \<rightarrow>\<^sub>\<beta> t'`
      have "{0 \<rightarrow>\<^sub>c [Fvar s,Fvar p]} E\<up>t \<rightarrow>\<^sub>\<beta> {0 \<rightarrow>\<^sub>c [Fvar s,Fvar p]} E\<up>t'" by blast
      with beta_lc[OF `t \<rightarrow>\<^sub>\<beta> t'`] sopen_and_copen
      have "{0 \<rightarrow> [Fvar s,Fvar p]} (E\<up>t) \<rightarrow>\<^sub>\<beta> {0 \<rightarrow> [Fvar s,Fvar p]} (E\<up>t')" by auto
      with sym[OF sclose_sopen_eq_t[OF `s \<notin> FV(E\<up>t')` `p \<notin> FV(E\<up>t')` `s \<noteq> p`]]
      show "(\<exists>t''. E\<up>t\<^bsup>[Fvar s,Fvar p]\<^esup> \<rightarrow>\<^sub>\<beta> t'' \<and> E\<up>t' = \<sigma>[s,p] t'')"
	by (rule_tac x = "{0 \<rightarrow> [Fvar s,Fvar p]} (E\<up>t')" in exI, 
	     simp add: openz_def closez_def)
    qed
  }
  note lem = this

(* induction *)
  show "lc_ctxt E \<Longrightarrow> \<forall>t t'. t \<rightarrow>\<^sub>\<beta> t' \<longrightarrow> E\<up>t \<rightarrow>\<^sub>\<beta> E\<up>t'"
  proof (induct E rule: lc_ctxt.induct, simp_all)
    case (lc_ctxt_Obj L f E T) 
    thus "\<forall>t t'. t \<rightarrow>\<^sub>\<beta> t' 
           \<longrightarrow> Obj (FLmap f(FLlabel f \<mapsto> E\<up>t)) T \<rightarrow>\<^sub>\<beta> Obj (FLmap f(FLlabel f \<mapsto> E\<up>t')) T"
    proof (intro strip)
      fix t :: sterm and t' :: sterm
      assume "t \<rightarrow>\<^sub>\<beta> t'"
      from `finite L` prems(2) have "\<forall>l\<in>dom (FLmap f). body (the((FLmap f) l))"
	unfolding body_def
	by (intro ballI, rule_tac x = L in exI, simp)
      from beta_Obj[OF _ _ lem[OF prems(3) `t \<rightarrow>\<^sub>\<beta> t'`] this] `finite L`
      show "Obj (FLmap f(FLlabel f \<mapsto> E\<up>t)) T \<rightarrow>\<^sub>\<beta> Obj (FLmap f(FLlabel f \<mapsto> E\<up>t')) T"
	by simp
    qed
  next
    case (lc_ctxt_UpdL E L u l)
    thus "\<forall>t t'. t \<rightarrow>\<^sub>\<beta> t' \<longrightarrow> Upd (E\<up>t) l u \<rightarrow>\<^sub>\<beta> Upd (E\<up>t') l u"
    proof (intro strip)
      fix t :: sterm and t' :: sterm
      assume "t \<rightarrow>\<^sub>\<beta> t'"
      from `finite L` prems(4) have "body u"
	unfolding body_def by (rule_tac x = L in exI, simp)
      from `\<forall>t t'. t \<rightarrow>\<^sub>\<beta> t' \<longrightarrow> E\<up>t \<rightarrow>\<^sub>\<beta> E\<up>t'` `t \<rightarrow>\<^sub>\<beta> t'` 
	   beta_UpdL[OF _ this]
      show "Upd (E\<up>t) l u \<rightarrow>\<^sub>\<beta> Upd (E\<up>t') l u" by simp
    qed
  next
    case (lc_ctxt_UpdR u L E l)
    thus "\<forall>t t'. t \<rightarrow>\<^sub>\<beta> t' \<longrightarrow> Upd u l (E\<up>t) \<rightarrow>\<^sub>\<beta> Upd u l (E\<up>t')"
    proof (intro strip)
      fix t :: sterm and t' :: sterm
      assume "t \<rightarrow>\<^sub>\<beta> t'"
      from `finite L` beta_UpdR[OF _ lem[OF prems(3) `t \<rightarrow>\<^sub>\<beta> t'`]] `lc u`
      show "Upd u l (E\<up>t) \<rightarrow>\<^sub>\<beta> Upd u l (E\<up>t')" by simp
    qed
  qed
qed

lemma sopen_context_Ref[rule_format]: 
  "\<forall>(E::general_context) s p n r. isref r 
    \<longrightarrow> {n \<rightarrow> [s,p]} t = E\<up>r 
    \<longrightarrow> (\<exists>E'::general_context. s = E'\<up>r)
       \<or>(\<exists>E'::general_context. p = E'\<up>r)
       \<or>(\<exists>E'::general_context. t = E'\<up>r)"
proof -
  have 
    "(\<forall>(E::general_context) s p n r. isref r
       \<longrightarrow> {n \<rightarrow> [s,p]} t = E\<up>r 
       \<longrightarrow> (\<exists>E'::general_context. s = E'\<up>r)
          \<or>(\<exists>E'::general_context. p = E'\<up>r)
          \<or>(\<exists>E'::general_context. t = E'\<up>r))
    \<and>(\<forall>(E::general_context) s p n r. isref r
       \<longrightarrow> sopen_option n s p u = Some (E\<up>r)
       \<longrightarrow> (\<exists>E'::general_context. s = E'\<up>r)
          \<or>(\<exists>E'::general_context. p = E'\<up>r)
          \<or>(\<exists>E'::general_context. u = Some (E'\<up>r)))"
  proof (induct rule: sterm.induct)
    case (Bvar b) thus ?case
      by (cases b, (case_tac "nat = n", auto)+)
  next
    case Fvar thus ?case by auto
  next
    case ActRef thus ?case by auto
  next
    case FutRef thus ?case by auto
  next
    case None_sterm thus ?case by auto
  next
    case Some_sterm thus ?case by auto    
  next
    case (Active a) thus ?case
    proof (intro strip, simp)
      fix r n f s p and E ::general_context
      assume "isref r" "Active ({n \<rightarrow> [s,p]} a) = E\<up>r"
      with ref_context_active[OF this(1) sym[OF this(2)]]
      obtain E1::general_context where "{n \<rightarrow> [s,p]} a = E1\<up>r" by auto
      with `isref r` prems(1) obtain E2::general_context 
	where "s = E2\<up>r \<or> p = E2\<up>r \<or> (\<exists>E::general_context. a = E\<up>r)"
	by blast
      thus "(\<exists>E'::general_context. s = E'\<up>r)
	   \<or>(\<exists>E'::general_context. p = E'\<up>r)
	   \<or>(\<exists>E'::general_context. Active a = E'\<up>r)"
      proof (elim disjE, blast, blast)
	assume "\<exists>E::general_context. a = E\<up>r" 
	from context_active_inverse[OF this] show ?thesis by simp
      qed
    qed
  next
    case (Call t1 l t2) thus ?case
    proof (intro strip, simp)
      fix r n f s p and E::general_context
      assume "isref r" "Call ({n \<rightarrow> [s,p]} t1) l ({n \<rightarrow> [s,p]} t2) = E\<up>r"
      from ref_context_call[OF this(1) sym[OF this(2)]] 
      obtain E1::general_context 
	where "{n \<rightarrow> [s,p]} t1 = E1\<up>r \<or> {n \<rightarrow>[s,p]} t2 = E1\<up>r" 
	by auto
      thus "(\<exists>E'::general_context. s = E'\<up>r)
	   \<or>(\<exists>E'::general_context. p = E'\<up>r)
	   \<or>(\<exists>E'::general_context. Call t1 l t2 = E'\<up>r)"
      proof (elim disjE)
	assume "{n \<rightarrow> [s,p]} t1 = E1\<up>r"
	with `isref r` prems(1) obtain E2::general_context
	where "s = E2\<up>r \<or> p = E2\<up>r \<or> t1 = E2\<up>r"
	  by blast
	thus ?thesis
	proof (elim disjE, blast, blast)
	  assume "t1 = E2\<up>r" with context_call_inverse[of t1 r t2]
	  show ?thesis by blast
	qed
      next
	assume "{n \<rightarrow> [s,p]} t2 = E1\<up>r" 
	with `isref r` prems(2) obtain E2::general_context
	where "s = E2\<up>r \<or> p = E2\<up>r \<or> t2 = E2\<up>r"
	  by blast
	thus ?thesis
	proof (elim disjE, blast, blast)
	  assume "t2 = E2\<up>r" with context_call_inverse[of t1 r t2]
	  show ?thesis by blast
	qed
      qed
    qed
  next
    case (Upd t1 l t2) thus ?case
    proof (intro strip, simp)
      fix r n f s p and E::general_context
      assume "isref r" "Upd ({n \<rightarrow> [s,p]} t1) l ({Suc n \<rightarrow> [s,p]} t2) = E\<up>r"
      from ref_context_upd[OF this(1) sym[OF this(2)]] 
      obtain E1::general_context 
	where "{n \<rightarrow> [s,p]} t1 = E1\<up>r \<or> {Suc n \<rightarrow>[s,p]} t2 = E1\<up>r" 
	by auto
      thus "(\<exists>E'::general_context. s = E'\<up>r)
	   \<or>(\<exists>E'::general_context. p = E'\<up>r)
	   \<or>(\<exists>E'::general_context. Upd t1 l t2 = E'\<up>r)"
      proof (elim disjE)
	assume "{n \<rightarrow> [s,p]} t1 = E1\<up>r" 
	with `isref r` prems(1) obtain E2::general_context
	where "s = E2\<up>r \<or> p = E2\<up>r \<or> t1 = E2\<up>r"
	  by blast
	thus ?thesis
	proof (elim disjE, blast, blast)
	  assume "t1 = E2\<up>r" with context_upd_inverse[of t1 r t2]
	  show ?thesis by blast
	qed
      next
	assume "{Suc n \<rightarrow> [s,p]} t2 = E1\<up>r" 
	with `isref r` prems(2) obtain E2::general_context
	where "s = E2\<up>r \<or> p = E2\<up>r \<or> t2 = E2\<up>r"
	  by blast
	thus ?thesis
	proof (elim disjE, blast, blast)
	  assume "t2 = E2\<up>r" with context_upd_inverse[of t1 r t2]
	  show ?thesis by blast
	qed
      qed
    qed
  next
    case (Obj f' T) thus ?case
    proof (intro strip, simp)
      fix r n f s p and E::general_context
      assume "isref r" "Obj (\<lambda>l. sopen_option (Suc n) s p (f' l)) T = E\<up>r"
      from ref_context_obj[OF this(1) sym[OF this(2)]] 
      obtain l and E1::general_context 
	where "sopen_option (Suc n) s p (f' l) = Some (E1\<up>r)" 
	by auto
      with `isref r` prems(1) obtain E2::general_context
	where "s = E2\<up>r \<or> p = E2\<up>r \<or> f' l = Some (E2\<up>r)"
	by blast
      thus "(\<exists>E'::general_context. s = E'\<up>r)
	   \<or>(\<exists>E'::general_context. p = E'\<up>r)
	   \<or>(\<exists>E'::general_context. Obj f' T = E'\<up>r)"
      proof (elim disjE, blast, blast)
	assume "f' l = Some (E2\<up>r)" with context_obj_inverse[of f' r T]
	show ?thesis by blast
      qed
    qed
  qed
  from conjunct1[OF this] show ?thesis by assumption
qed

end
